//==================================================================
//
//	File name	:	list.js
//
//	Summary		:	ListBox ComboBox classes
//
//	Ceated		:	2018/11/25
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (list) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	list.STYLE_DEFAULT = 0;
	list.STYLE_CHECKBOX = 1;
	list.STYLE_POPUP = 2;
	list.STYLE_EDITABLE = 2;
	list.SORT_BY_TEXT = 0;
	list.SORT_BY_VALUE = 1;
	list.SORT_ASC = 0;
	list.SORT_DESC = 1;
	
	// Private variables
	let oDropDown = null;
	let bDropDownKeep = false;
	let oCurrentMouseDown = null;
	
	
	//==========================================================
	//
	//	Class		:	ListItem
	//
	//	Summary		:	Class of list item
	//
	//==========================================================
	list.ListItem = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(oParent,		// Parent control[ListBox or ComboBox]
					iStyle,			// Style of list[int]
					iWidth,			// Width[int]
					iAlign,			// Text align
					sText,			// Text to display[string]
					sVal,			// Value of this item[string]
					bEnabled,		// Enabled or not[boolean]
					bDraggable,		// Draggable or not[boolean]
					fpDragstart)	// Function on start dragging[function]
		{
			// Check parameters
			if((typeof oParent != "object")
			|| ((!oParent instanceof list.ListBox) && (!oParent instanceof list.ComboBox)))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_PARENT, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			
			if((iStyle != list.STYLE_DEFAULT)
			&& (iStyle != list.STYLE_CHECKBOX)
			&& (iStyle != list.STYLE_POPUP)
			&& (iStyle != list.STYLE_EDITABLE))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			
			if(typeof iWidth != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WIDTH, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			if(iWidth < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_WIDTH, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			
			if(typeof sText != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			
			if(typeof sVal != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			
			if(typeof bEnabled != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ENABLED, ""),
											"list.js", "jaswin.list.ListItem.constructor");
			
			if(bDraggable)
			{
				if(typeof bDraggable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DRAGGABLE, ""),
												"list.js", "jaswin.list.ListItem.constructor");
				if(fpDragstart)
				{
					if(typeof fpDragstart != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DRAGSTART, ""),
													"list.js", "jaswin.list.ListItem.constructor");
				}
			}
			
			// Set properties
			this.parent = oParent;			// Parent control[ListBox or ComboBox]
			this.style = iStyle;			// Style of list[int]
			this.value = sVal;				// Value of this item[string]
			this.text = sText;				// Text to display[string]
			this.selected = false;			// Selected or not
			this.element = null;			// Element[DOM Element]
			this.box = null;				// Element for checkbox[DOM Element]
			this.label = null;				// Element for text label[DOM Element]
			this.draggable = bDraggable;	// Draggable or not[boolean]
			this.onDragstart = fpDragstart;	// Function on dragstart[function(Argument:ListItem class, event)]
			
			// Create element
			this.element = document.createElement("li");
			if(bEnabled)
				this.element.className = "list-item";
			else
				this.element.className = "list-item-disabled";
			
			if(this.style == list.STYLE_CHECKBOX)
			{
				// Checkbox
				this.box = document.createElement("span");
				this.box.className = jaswin.ICON_CHECKBOX;
				this.box.style.left = "2px";
				this.box.style.top = "3px";
				this.element.appendChild(this.box);
				if(!bEnabled)
					oItem.box.style.cursor = "default";
			}
			// Label
			this.label = document.createElement("span");
			if(bEnabled)
				this.label.className = "list-item-label";
			else
				this.label.className = "list-item-label-disabled";
			
			if(iAlign == jaswin.ALIGN_RIGHT)
				this.label.style.textAlign = "right";
			else if(iAlign == jaswin.ALIGN_CENTER)
				this.label.style.textAlign = "center";
			
			if(this.style == list.STYLE_CHECKBOX)
				this.label.style.left = "18px";
			else
				this.label.style.left = "2px";
			this.label.style.width = iWidth + "px";
			this.label.innerHTML = this.text;
			this.element.appendChild(this.label);
			
			if(bEnabled)
			{
				this.element.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
					this.parent.setHoverList(this);
				}, false);
				this.element.addEventListener("mouseout", this.mouseOutEvent = (e) => {
					this.parent.setOutList(this);
				}, false);
				this.element.addEventListener("mousedown", this.mouseDownEvent = (e) => {
					this.parent.clickList(this);
				}, false);
				
				if(this.draggable)
				{
					this.element.draggable = true;
					this.element.addEventListener("dragstart", this.dragStartEvent = (e) => {
						this.onDragstart(this, e);
					}, false);
				}
			}
		}
		
		//==================================================
		//	Enable this item
		//==================================================
		enable()
		{
			if((!this.element) || (!this.parent))
				return;
			
			if(this.selected)
			{
				this.element.className = "list-item-selected";
				if(this.style == list.STYLE_CHECKBOX)
					this.box.className = jaswin.ICON_CHECKBOX_CHECKED;
			}
			else
				this.element.className = "list-item";
			
			if(this.style == list.STYLE_CHECKBOX)
				this.box.style.cursor = "pointer";
			this.label.className = "list-item-label";
			
			if(!this.mouseMoveEvent)
			{
				this.element.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
					this.parent.setHoverList(this);
				}, false);
			}
			if(!this.mouseOutEvent)
			{
				this.element.addEventListener("mouseout", this.mouseOutEvent = (e) => {
					this.parent.setOutList(this);
				}, false);
			}
			if(!this.mouseDownEvent)
			{
				this.element.addEventListener("mousedown", this.mouseDownEvent = (e) => {
					this.parent.clickList(this);
				}, false);
			}
			
			if(this.draggable)
			{
				if(!this.dragStartEvent)
				{
					this.element.draggable = true;
					this.element.addEventListener("dragstart", this.dragStartEvent = (e) => {
						this.onDragstart(this, e);
					}, false);
				}
			}
		}
		
		//==================================================
		//	Disable this item
		//==================================================
		disable()
		{
			if((!this.element) || (!this.parent))
				return;
			
			if(this.selected)
			{
				this.element.className = "list-item-selected-disabled";
				if(this.style == list.STYLE_CHECKBOX)
					this.box.className = jaswin.ICON_CHECKBOX_CHECKED_DISABLED;
			}
			else
				this.element.className = "list-item-disabled";
			
			if(this.style == list.STYLE_CHECKBOX)
				this.box.style.cursor = "default";
			this.label.className = "list-item-label-disabled";
			
			if(this.mouseMoveEvent)
			{
				this.element.removeEventListener("mousemove", this.mouseMoveEvent, false);
				this.mouseMoveEvent = null;
			}
			if(this.mouseOutEvent)
			{
				this.element.removeEventListener("mouseout", this.mouseOutEvent, false);
				this.mouseOutEvent = null;
			}
			if(this.mouseDownEvent)
			{
				this.element.removeEventListener("mousedown", this.mouseDownEvent, false);
				this.mouseDownEvent = null;
			}
			
			if(this.draggable)
			{
				this.element.draggable = false;
				if(this.dragStartEvent)
				{
					this.element.removeEventListener("dragstart", this.dragStartEvent, false);
					this.dragStartEvent = null;
				}
			}
		}
	}
	
	//==========================================================
	//
	//	Class		:	ListBox
	//
	//	Summary		:	Class of listbox
	//
	//==========================================================
	list.ListBox = class extends jaswin.Control
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[string]
					oRect,			// * Size and position[FixedRect]
					iStyle,			// * Item style[int]
					bEnabled,		// * Enabled or not[boolean]
					sTooltip,		// Tool tip text[string]
					iTabIndex,		// Tab index[int]
					iAlign,			// Text align[int]
					fpChange,		// Function on change[function](ListItem)
					bDraggable,		// Draggable or not[boolean]
					fpDragstart)	// Function on start dragging[function](ListItem, event)
		{
			// Base class
			super(false, sID, false, oRect, bEnabled, iTabIndex);
			
			// Check parameters
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"list.js", "jaswin.list.ListBox.constructor");
			
			if((iStyle != list.STYLE_DEFAULT)
			&& (iStyle != list.STYLE_CHECKBOX)
			&& (iStyle != list.STYLE_POPUP))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, sID),
											"list.js", "jaswin.list.ListBox.constructor");
			
			if(sTooltip)
			{
				if(typeof sTooltip != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, sID),
												"button.js", "jaswin.button.Button.constructor");
			}
			
			if(iAlign)
			{
				if(typeof iAlign != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ALIGN, sID),
												"list.js", "jaswin.list.ListBox.constructor");
				if((iAlign != jaswin.ALIGN_LEFT)
				&& (iAlign != jaswin.ALIGN_RIGHT)
				&& (iAlign != jaswin.ALIGN_CENTER))
					throw new jaswin.JaswinError(jaswin.ALERT_VALUE_ALIGN,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ALIGN, sID),
												"list.js", "jaswin.list.ListBox.constructor");
			}
			
			if(fpChange)
			{
				if(typeof fpChange != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, sID),
												"list.js", "jaswin.list.ListBox.constructor");
			}
			
			if(bDraggable)
			{
				if(typeof bDraggable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DRAGGABLE, sID),
												"list.js", "jaswin.list.ListBox.constructor");
				if(fpDragstart)
				{
					if(typeof fpDragstart != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DRAGSTART, sID),
													"list.js", "jaswin.list.ListBox.constructor");
				}
			}
			
			// Set properties
			this.style = iStyle;				// ListBox style[int]
			this.onChange = fpChange;			// Function on change[function](ListItem)
			this.items = new Array();			// List items[ListItem]
			this.selectedIndex = -1;			// Selected index[int]
			this.selectedIndexes = new Array();	// Selected indexes[Array of int]
			this.currentIndex = -1;				// Current index of item on focus[int]
			this.draggable = bDraggable;		// Draggable or not[boolean]
			this.onDragstart = fpDragstart;		// Function on dragstart[function(Argument:ListItem class, event)]
			if(iAlign)							// Text align of label[int]
				this.align = iAlign;
			else
				this.align = jaswin.ALIGN_LEFT;
			
			// Create element
			this.element = document.createElement("ul");
			if(this.style == list.STYLE_POPUP)
				this.element.className = "listbox-popup";
			else
				this.element.className = "listbox";
			
			if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
				this.element.style.right = this.rect.posX + "px";
			else
				this.element.style.left = this.rect.posX + "px";
			
			if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
				this.element.style.bottom = this.rect.posY + "px";
			else
				this.element.style.top = this.rect.posY + "px";
			
			this.element.style.width = this.rect.width - 2 + "px";		// 2:border
			this.element.style.height = this.rect.height - 2 + "px";	// 2:border
			this.focusElement = this.element;
			
			// Tooltip
			if((sTooltip) && (sTooltip.length > 0))
				this.element.title = sTooltip;
			
			// Tab index
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				this.element.addEventListener("keydown", this.keyUpEvent = (e) => {
					this.keyDownList(e);
				}, false);
				this.element.addEventListener("blur", this.blurEvent = (e) => {
					this.blur();
				}, false);
				
				if(this.style == list.STYLE_POPUP)
				{
					global.element.addEventListener("click", this.cancelEvent = (e) => {
						this.hide();
					}, false);
				}
			}
		}
		
		//==================================================
		//	Set this control to parent class
		//==================================================
		setParent(oParent)
		{
			super.setParent(oParent);
			
			let oParentWnd = jaswin.findParentWindow(this);
			
			if((jaswin.frame) && (oParentWnd != global))
			{
				if(this.style == list.STYLE_POPUP)
					this.element.style.zIndex = 50 + jaswin.frame.windowCount + oParentWnd.modalLevel;
			}
		}
		
		//==================================================
		//	Enable this listbox
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				if(!this.keyUpEvent)
				{
					this.element.addEventListener("keydown", this.keyUpEvent = (e) => {
						this.keyDownList(e);
					}, false);
				}
				if(!this.blurEvent)
				{
					this.element.addEventListener("blur", this.blurEvent = (e) => {
						this.blur();
					}, false);
				}
				if((this.style == list.STYLE_POPUP) && (this.element.style.display != "none"))
				{
					if(!this.cancelEvent)
					{
						global.element.addEventListener("click", this.cancelEvent = (e) => {
							this.hide();
						}, false);
					}
				}
				
				let iCounter;
				for(iCounter = 0; iCounter < this.items.length; iCounter++)
					this.items[iCounter].enable();
			}
		}
		
		//==================================================
		//	Disable this listbox
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			if(this.keyUpEvent)
			{
				this.element.removeEventListener("keydown", this.keyUpEvent, false);
				this.keyUpEvent = null;
			}
			if(this.blurEvent)
			{
				this.element.removeEventListener("blur", this.blurEvent, false);
				this.blurEvent = null;
			}
			if((this.style == list.STYLE_POPUP) && (this.element.style.display != "none"))
			{
				if(this.cancelEvent)
				{
					global.element.removeEventListener("click", this.cancelEvent, false);
					this.cancelEvent = null;
				}
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
				this.items[iCounter].disable();
		}
		
		//==================================================
		//	Show this listbox
		//==================================================
		show()
		{
			super.show();
			
			if((this.style == list.STYLE_POPUP) && (this.enabled))
			{
				if(!this.cancelEvent)
				{
					global.element.addEventListener("click", this.cancelEvent = (e) => {
						this.hide();
					}, false);
				}
			}
		}
		
		//==================================================
		//	Hide this listbox
		//==================================================
		hide()
		{
			super.hide();
			
			if((this.style == list.STYLE_POPUP) && (this.enabled))
			{
				if(this.cancelEvent)
				{
					global.element.removeEventListener("click", this.cancelEvent, false);
					this.cancelEvent = null;
				}
			}
		}
		
		//==================================================
		//	Add item to this listbox
		//==================================================
		addItem(sText, sVal, bSelected)
		{
			if(typeof sText != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, this.id),
											"list.js", "jaswin.list.ListBox.addItem");
			
			if(typeof sVal != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
											"list.js", "jaswin.list.ListBox.addItem");
			
			if(typeof bSelected != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SELECTED, this.id),
											"list.js", "jaswin.list.ListBox.addItem");
			
			let iWidth;
			let iCounter;
			let oItem;
			let bEnabled = false;
			
			// Adjust list width and height
			iWidth = this.rect.width - 2 - (1 * 2) - (1 * 2);	// 2:padding 1*2:border(frame) 1*2:border(item)
			if(this.style == list.STYLE_CHECKBOX)
				iWidth -= 17;									// 17:checkbox
			
			if(this.style == list.STYLE_POPUP)
			{
				// Height when popup
				if((this.items.length + 1) * 20 < this.rect.height)
					this.element.style.height = (this.items.length + 1) * 20 + "px";
				else
					this.element.style.height = this.rect.height + "px";
			}
			
			if(((this.items.length + 1) * 20) > this.rect.height - 2)
			{
				// Width of label
				iWidth -= 17;									// 17:scrollbar
				for(iCounter = 0; iCounter < this.items.length; iCounter++)
					this.items[iCounter].label.style.width = iWidth + "px";
			}
			
			// Make a new item
			if((this.enabled) && (this.parentEnabled))
				bEnabled = true;
			oItem = new list.ListItem(this, this.style, iWidth, this.align,
										sText, sVal, bEnabled, this.draggable, this.onDragstart);
			
			this.element.appendChild(oItem.element);
			this.items.push(oItem);
			
			if(bSelected)
				this.selectList(this.items.length - 1, false);
		}
		
		//==================================================
		//	Remove item from this listbox
		//==================================================
		removeItem(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"list.js", "jaswin.list.ListBox.removeItem");
			if((iIndex < 0) || (iIndex >= this.items.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"list.js", "jaswin.list.ListBox.removeItem");
			
			// Remove an item
			this.element.removeChild(this.items[iIndex].element);
			this.items.splice(iIndex, 1);
			
			// Adjust list width and height
			let iWidth;
			iWidth = this.rect.width - 2 - (1 * 2) - (1 * 2);	// 2:padding 1*2:border(frame) 1*2:border(item)
			if(this.style == list.STYLE_CHECKBOX)
				iWidth -= 17;									// 17:checkbox
			
			if((this.items.length * 20) > this.rect.height - 2)
				iWidth -= 17;									// 17:scrollbar
			
			if(this.style == list.STYLE_POPUP)
			{
				// Height when popup
				if((this.items.length - 1) * 20 < this.rect.height)
					this.element.style.height = (this.items.length - 1) * 20 + "px";
				else
					this.element.style.height = this.rect.height + "px";
			}
			
			// Width of label and selection
			this.selectedIndexes = new Array();
			this.selectedIndex = -1;
			this.currentIndex = -1;
			
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				this.items[iCounter].label.style.width = iWidth + "px";
				if(this.items[iCounter].selected)
				{
					if(this.style == list.STYLE_CHECKBOX)
						this.selectedIndexes.push(iCounter);
					else
						this.selectedIndexes[0] = iCounter;
					this.selectedIndex = iCounter;
				}
			}
		}
		
		//==================================================
		//	Set list item of this listbox hover
		//==================================================
		setHoverList(oItem)
		{
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
				{
					this.currentIndex = iCounter;
					if(!oItem.selected)
					{
						oItem.element.className = "list-item-hover";
						if(this.style == list.STYLE_CHECKBOX)
							oItem.box.className = jaswin.ICON_CHECKBOX_HOVER;
					}
					break;
				}
			}
		}
		
		//==================================================
		//	Set list item of this listbox mouse out
		//==================================================
		setOutList(oItem)
		{
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
				{
					this.currentIndex = -1;
					if(!oItem.selected)
					{
						oItem.element.className = "list-item";
						if(this.style == list.STYLE_CHECKBOX)
							oItem.box.className = jaswin.ICON_CHECKBOX;
					}
					break;
				}
			}
		}
		
		//==================================================
		//	Click list item of this listbox
		//==================================================
		clickList(oItem)
		{
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
				{
					this.selectList(iCounter, true);
					break;
				}
			}
		}
		
		//==================================================
		//	Set list item of this listbox focused
		//==================================================
		setFocusItem(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"list.js", "jaswin.list.ListBox.setFocusItem");
			
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(iCounter == iIndex)
				{
					this.currentIndex = iCounter;
					if(!this.items[iCounter].selected)
					{
						this.items[iCounter].element.className = "list-item-hover";
						if(this.style == list.STYLE_CHECKBOX)
							this.items[iCounter].box.className = jaswin.ICON_CHECKBOX_HOVER;
					}
				}
				else
				{
					if(!this.items[iCounter].selected)
					{
						this.items[iCounter].element.className = "list-item";
						if(this.style == list.STYLE_CHECKBOX)
							this.items[iCounter].box.className = jaswin.ICON_CHECKBOX;
					}
				}
			}
			
			if(iIndex != -1)
			{
				if(iIndex * 20 < this.element.scrollTop)
					this.element.scrollTop = iIndex * 20;
				else if((iIndex + 1) * 20 > (this.rect.height - 2) + this.element.scrollTop)
					this.element.scrollTop = (iIndex + 1) * 20 - ((this.rect.height - 2) + this.element.scrollTop);
			}
		}
		
		//==================================================
		//	Select list item of this listbox
		//==================================================
		selectList(iIndex, bExecFunc)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"list.js", "jaswin.list.ListBox.selectList");
			if((iIndex < 0) || (iIndex >= this.items.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"list.js", "jaswin.list.ListBox.selectList");
			
			let bExec = true;
			
			if((typeof bExecFunc != "undefined") && (bExecFunc != null))
			{
				if(typeof bExecFunc != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_EXEC_FUNC, this.id),
												"list.js", "jaswin.list.ListBox.selectList");
				bExec = bExecFunc;
			}
			
			if((this.style != list.STYLE_CHECKBOX)
			&& (iIndex == this.selectedIndex))
				return;
			
			let iCounter;
			
			this.selectedIndexes = new Array();
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.style == list.STYLE_CHECKBOX)
				{
					if(iCounter == iIndex)
					{
						if(this.items[iIndex].selected)
						{
							if((this.enabled) && (this.parentEnabled))
							{
								this.items[iIndex].element.className = "list-item-hover";
								this.items[iIndex].box.className = jaswin.ICON_CHECKBOX_HOVER;
							}
							else
							{
								this.items[iIndex].element.className = "list-item-disabled";
								this.items[iIndex].box.className = jaswin.ICON_CHECKBOX;
							}
							this.items[iIndex].selected = false;
						}
						else
						{
							if((this.enabled) && (this.parentEnabled))
							{
								this.items[iIndex].element.className = "list-item-selected";
								this.items[iIndex].box.className = jaswin.ICON_CHECKBOX_CHECKED;
							}
							else
							{
								this.items[iIndex].element.className = "list-item-selected-disabled";
								this.items[iIndex].box.className = jaswin.ICON_CHECKBOX_CHECKED_DISABLED;
							}
							this.items[iIndex].selected = true;
							this.selectedIndexes.push(iIndex);
						}
					}
					else if(this.items[iCounter].selected)
						this.selectedIndexes.push(iCounter);
				}
				else
				{
					if(iCounter == iIndex)
					{
						if((this.enabled) && (this.parentEnabled))
							this.items[iIndex].element.className = "list-item-selected";
						else
							this.items[iIndex].element.className = "list-item-selected-disabled";
						this.items[iIndex].selected = true;
						this.selectedIndex = iIndex;
					}
					else
					{
						if((this.enabled) && (this.parentEnabled))
							this.items[iCounter].element.className = "list-item";
						else
							this.items[iCounter].element.className = "list-item-disabled";
						this.items[iCounter].selected = false;
					}
				}
			}
			
			if(this.style != list.STYLE_CHECKBOX)
			{
				this.selectedIndex = iIndex;
				this.selectedIndexes[0] = iIndex;
			}
			
			if((this.onChange) && (this.enabled) && (this.parentEnabled)
			&& (bExec) && (iIndex != -1))
				this.onChange(this.items[iIndex]);
		}
		
		//==================================================
		//	Key down on this listbox
		//==================================================
		keyDownList(e)
		{
			if((e.keyCode == 13) || (e.keyCode == 32))
			{
				if(this.currentIndex != -1)
				{
					this.selectList(this.currentIndex, true);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			}
			else if(e.keyCode == 40)
			{
				// down
				if(this.currentIndex < this.items.length - 1)
					this.setFocusItem(this.currentIndex + 1);
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			else if(e.keyCode == 38)
			{
				// up
				if(this.currentIndex > 0)
					this.setFocusItem(this.currentIndex - 1);
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			return true;
		}
		
		//==================================================
		//	Blur from this listbox
		//==================================================
		blur()
		{
			this.currentIndex = -1;
			this.setFocusItem(-1);
		}
		
		//==================================================
		//	Clear list of this listbox
		//==================================================
		clear()
		{
			this.selectedIndex = -1;
			this.selectedIndexes = new Array();
			this.items = new Array();
			this.element.innerHTML = "";
		}
		
		//==================================================
		//	Get selected items
		//==================================================
		getSelectedItems()
		{
			let oaItems;
			let iCounter;
			
			oaItems = new Array();
			for(iCounter = 0; iCounter < this.selectedIndexes.length; iCounter++)
				oaItems.push(this.items[this.selectedIndexes[iCounter]]);
			
			return oaItems;
		}
		
		//==================================================
		//	Get text by value
		//==================================================
		getTextByValue(sValue)
		{
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter].value == sValue)
					return this.items[iCounter].text;
			}
			
			return null;
		}
		
		//==================================================
		//	Contain item or not
		//==================================================
		contains(oItem)
		{
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
					return true;
			}
			
			return false;
		}
		
		//==================================================
		//	Sort
		//==================================================
		sort(iSortBy, iAscDesc)
		{
			if(typeof iSortBy != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SORT_BY, this.id),
											"list.js", "jaswin.list.ListBox.sort");
			if(typeof iAscDesc != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SORT_TYPE, this.id),
											"list.js", "jaswin.list.ListBox.sort");
			if(!this.items.length)
				return;
			
			// Execute sort
			if(iSortBy == list.SORT_BY_TEXT)
			{
				if(iAscDesc == list.SORT_ASC)
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.text < oItemB.text) return -1;
										if(oItemA.text > oItemB.text) return 1;
										return 0;
									});
				}
				else
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.text > oItemB.text) return -1;
										if(oItemA.text < oItemB.text) return 1;
										return 0;
									});
				}
			}
			else
			{
				if(iAscDesc == list.SORT_ASC)
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.value < oItemB.value) return -1;
										if(oItemA.value > oItemB.value) return 1;
										return 0;
									});
				}
				else
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.value > oItemB.value) return -1;
										if(oItemA.value < oItemB.value) return 1;
										return 0;
									});
				}
			}
			
			// Replace elements
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
				this.items[iCounter].element.parentNode.removeChild(this.items[iCounter].element);
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
				this.element.appendChild(this.items[iCounter].element);
		}
		
		//==================================================
		//	Set tooltip
		//==================================================
		setTooltip(sTooltip)
		{
			if(typeof sTooltip != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, this.id),
											"list.js", "jaswin.list.ListBox.setTooltip");
			this.element.title = sTooltip;
		}
		
		//==================================================
		//	Set change event
		//==================================================
		setChangeEvent(fpChange)
		{
			if(typeof fpChange != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, this.id),
											"list.js", "jaswin.list.ListBox.setChangeEvent");
			
			this.onChange = fpChange;
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	ComboBox
	//
	//	Summary		:	Class of combobox
	//
	//==========================================================
	list.ComboBox = class extends jaswin.Control
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[string]
					oRect,			// * Size and position[FixedRect]
					iStyle,			// * Item style[int]
					bEnabled,		// * Enabled or not[boolean]
					iMaxDropdown,	// Number of drop down items[int]
					sTooltip,		// Tool tip text[string]
					iTabIndex,		// Tab index[int]
					iAlign,			// Text align[int]
					fpChange,		// Function on change[function](ListItem)
					bDraggable,		// Draggable or not[boolean]
					fpDragstart)	// Function on start dragging[function](ListItem, event)
		{
			// Base class
			super(false, sID, false, oRect, bEnabled, iTabIndex);
			
			// Check parameters
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"list.js", "jaswin.list.ComboBox.constructor");
			
			if((iStyle != list.STYLE_DEFAULT)
			&& (iStyle != list.STYLE_CHECKBOX)
			&& (iStyle != list.STYLE_EDITABLE))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, sID),
											"list.js", "jaswin.list.ComboBox.constructor");
			
			if(iMaxDropdown)
			{
				if(typeof iMaxDropdown != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_MAX_DROP, sID),
												"list.js", "jaswin.list.ComboBox.constructor");
			}
			
			if(sTooltip)
			{
				if(typeof sTooltip != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, sID),
												"list.js", "jaswin.list.ComboBox.constructor");
			}
			
			if(iAlign)
			{
				if(typeof iAlign != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ALIGN, sID),
												"list.js", "jaswin.list.ComboBox.constructor");
				if((iAlign != jaswin.ALIGN_LEFT)
				&& (iAlign != jaswin.ALIGN_RIGHT)
				&& (iAlign != jaswin.ALIGN_CENTER))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_ALIGN, sID),
												"list.js", "jaswin.list.ComboBox.constructor");
			}
			
			if(fpChange)
			{
				if(typeof fpChange != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, sID),
												"list.js", "jaswin.list.ComboBox.constructor");
			}
			
			if(bDraggable)
			{
				if(typeof bDraggable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DRAGGABLE, sID),
												"list.js", "jaswin.list.ComboBox.constructor");
				if(fpDragstart)
				{
					if(typeof fpDragstart != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DRAGSTART, sID),
													"list.js", "jaswin.list.ComboBox.constructor");
				}
			}
			
			// Set properties
			this.style = iStyle;					// Combobox style[int]
			if(iMaxDropdown)
				this.maxDropdown = iMaxDropdown;	// Number of drop down items[int]
			else
				this.maxDropdown = 8;
			this.onChange = fpChange;				// Function on change[function](ListItem)
			this.box = null;						// Box element[DOM Element]
			this.text = null;						// Textbox or label element[DOM Element]
			this.arrow = null;						// Arrow button element[DOM Element]
			this.list = null;						// Content element for list[DOM Element]
			this.items = new Array();				// List items[Array of ListItem]
			this.dropDown = false;					// Drop down or not[boolean]
			this.selectedIndex = -1;				// Selected index[int]
			this.selectedIndexes = new Array();		// Selected indexes[Array of int]
			this.currentIndex = -1;					// Current index of item on focus[int]
			this.draggable = bDraggable;			// Draggable or not[boolean]
			this.onDragstart = fpDragstart;			// Function on dragstart[function(Argument:ListItem class, event)]
			if(iAlign)								// Text align of label[int]
				this.align = iAlign;
			else
				this.align = jaswin.ALIGN_LEFT;
			
			// Create element
			this.element = document.createElement("div");
			this.element.style.position = "absolute";
			
			if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
				this.element.style.right = this.rect.posX + "px";
			else
				this.element.style.left = this.rect.posX + "px";
			
			if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
				this.element.style.bottom = this.rect.posY + "px";
			else
				this.element.style.top = this.rect.posY + "px";
			this.element.style.width = this.rect.width + "px";
			this.element.style.height = "20px";
			
			// Box part
			this.box = document.createElement("div");
			this.box.className = "combobox";
			
			if(this.style == list.STYLE_EDITABLE)
			{
				// Editable
				this.text = document.createElement("input");
				this.text.type = "text";
				this.text.className = "combobox-text";
				this.text.style.width = this.rect.width - 2 - 3 - 1 - 17 + "px";	// 2:border 3:left 1:arrow border 17:arrow
				
				if(this.enabled)
					this.text.className = "combobox-text";
				else
					this.text.disabled = true;
				
				if(this.align == jaswin.ALIGN_RIGHT)
					this.text.style.textAlign = "right";
				else if(this.align == jaswin.ALIGN_CENTER)
					this.text.style.textAlign = "center";
				
				this.box.appendChild(this.text);
			}
			else
			{
				this.text = document.createElement("span");
				this.text.style.width = this.rect.width - 2 - 3 - 17 + "px";	// 2:border 3:left 17:arrow
				if(this.enabled)
					this.text.className = "combobox-label";
				else
					this.text.className = "combobox-label-disabled";
				
				if(this.align == jaswin.ALIGN_RIGHT)
					this.text.style.textAlign = "right";
				else if(this.align == jaswin.ALIGN_CENTER)
					this.text.style.textAlign = "center";
				
				this.box.appendChild(this.text);
			}
			
			let oTriangle;
			this.arrow = document.createElement("div");
			oTriangle = document.createElement("span");
			this.arrow.appendChild(oTriangle);
			if(this.enabled)
			{
				this.arrow.className = "combobox-arrow";
				oTriangle.className = jaswin.ICON_TRIANGLE_BOTTOM;
			}
			else
			{
				this.arrow.className = "combobox-arrow-disabled";
				oTriangle.className = jaswin.ICON_TRIANGLE_BOTTOM_DISABLED;
			}
			oTriangle.style.left = "4px";
			oTriangle.style.top = "6px";
			
			this.box.appendChild(this.arrow);
			this.element.appendChild(this.box);
			
			// List part
			this.list = document.createElement("ul");
			this.list.className = "combobox-list";
			this.list.style.width = this.rect.width - 2 + "px";				// 2:border
			this.list.style.height = "0px";
			this.element.appendChild(this.list);
			
			// Tooltip
			if((sTooltip) && (sTooltip.length > 0))
				this.element.title = sTooltip;
			
			// Tab index
			if(this.style == list.STYLE_EDITABLE)
				this.focusElement = this.text;
			else
				this.focusElement = this.element;
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				if(this.style == list.STYLE_EDITABLE)
				{
					this.text.addEventListener("focus", this.focusEvent = (e) => {
						this.focusText();
					}, false);
					this.text.addEventListener("blur", this.blurEvent = (e) => {
						this.blurText();
					}, false);
					this.text.addEventListener("input", this.inputEvent = (e) => {
						this.inputText();
					}, false);
				}
				else
				{
					this.element.addEventListener("focus", this.focusEvent = (e) => {
						if((!this.enabled) || (!this.parentEnabled))
						{
							e.preventDefault();
							return;
						}
						this.setHover();
					}, false);
					this.element.addEventListener("blur", this.blurEvent = (e) => {
						if(bDropDownKeep)
						{
							bDropDownKeep = false;
							return;
						}
						
						if(this == oCurrentMouseDown)
							return;
						
						if((!this.enabled) || (!this.parentEnabled))
						{
							e.preventDefault();
							return;
						}
						this.blur();
					}, false);
					this.text.addEventListener("click", this.clickEvent = (e) => {
						if((oDropDown) && (oDropDown != this))
							oDropDown.click();
						this.click();
					}, false);
				}
				
				this.arrow.addEventListener("click", this.arrowClickEvent = (e) => {
					if((oDropDown) && (oDropDown != this))
						oDropDown.click();
					this.click();
				}, false);
				this.element.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
					this.setHover();
				}, false);
				this.element.addEventListener("mouseout", this.mouseOutEvent = (e) => {
					this.setOut();
				}, false);
				this.element.addEventListener("keydown", this.keyDownEvent = (e) => {
					return this.keyDownCombo(e);
				}, false);
				this.list.addEventListener("mousedown", this.mouseDownEvent = (e) => {
					let oElement = e.target;
					let iCounter;
					
					for(iCounter = 0; iCounter < this.items.length; iCounter++)
					{
						if(this.items[iCounter].element.contains(oElement))
							return;
					}
					oCurrentMouseDown = this;
				}, false);
				this.list.addEventListener("mouseup", this.mouseUpEvent = (e) => {
					oCurrentMouseDown = null;
				}, false);
			}
		}
		
		//==================================================
		//	Enable this combobox
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				this.arrow.className = "combobox-arrow";
				this.arrow.childNodes[0].className = jaswin.ICON_TRIANGLE_BOTTOM;
				if(!this.arrowClickEvent)
				{
					this.arrow.addEventListener("click", this.arrowClickEvent = (e) => {
						if((oDropDown) && (oDropDown != this))
							oDropDown.click();
						this.click();
					}, false);
				}
				
				if(this.style == list.STYLE_EDITABLE)
				{
					this.text.className = "combobox-text";
					this.text.disabled = false;
					if(!this.focusEvent)
					{
						this.text.addEventListener("focus", this.focusEvent = (e) => {
							this.focusText();
						}, false);
					}
					if(!this.blurEvent)
					{
						this.text.addEventListener("blur", this.blurEvent = (e) => {
							this.blurText();
						}, false);
					}
					if(!this.inputEvent)
					{
						this.text.addEventListener("input", this.inputEvent = (e) => {
							this.inputText();
						}, false);
					}
					if(!this.comboArrowDown)
					{
						this.arrow.addEventListener("mousedown", this.comboArrowDown = (e) => {
							if((oDropDown) && (oDropDown == this))
								this.closingList = true;
						}, false);
					}
				}
				else
				{
					this.text.className = "combobox-label";
					if(!this.focusEvent)
					{
						this.element.addEventListener("focus", this.focusEvent = (e) => {
							if((!this.enabled) || (!this.parentEnabled))
							{
								e.preventDefault();
								return;
							}
							this.setHover();
						}, false);
					}
					if(!this.blurEvent)
					{
						this.element.addEventListener("blur", this.blurEvent = (e) => {
							if(bDropDownKeep)
							{
								bDropDownKeep = false;
								return;
							}
							
							if(this == oCurrentMouseDown)
								return;
							
							if((!this.enabled) || (!this.parentEnabled))
							{
								e.preventDefault();
								return;
							}
							this.blur();
						}, false);
					}
					if(!this.clickEvent)
					{
						this.text.addEventListener("click", this.clickEvent = (e) => {
							if((oDropDown) && (oDropDown != this))
								oDropDown.click();
							this.click();
						}, false);
					}
				}
				if(!this.mouseMoveEvent)
				{
					this.element.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
						this.setHover();
					}, false);
				}
				if(!this.mouseOutEvent)
				{
					this.element.addEventListener("mouseout", this.mouseOutEvent = (e) => {
						this.setOut();
					}, false);
				}
				if(!this.keyDownEvent)
				{
					this.element.addEventListener("keydown", this.keyDownEvent = (e) => {
						return this.keyDownCombo(e);
					}, false);
				}
				if(!this.mouseDownEvent)
				{
					this.list.addEventListener("mousedown", this.mouseDownEvent = (e) => {
						let oElement = e.target;
						let iCounter;
						
						for(iCounter = 0; iCounter < this.items.length; iCounter++)
						{
							if(this.items[iCounter].element.contains(oElement))
								return;
						}
						oCurrentMouseDown = this;
					}, false);
				}
				if(!this.mouseUpEvent)
				{
					this.list.addEventListener("mouseup", this.mouseUpEvent = (e) => {
						oCurrentMouseDown = null;
					}, false);
				}
			}
		}
		
		//==================================================
		//	Disable this combobox
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			this.list.style.display = "none";
			this.dropDown = false;
			this.box.className = "combobox";
			oDropDown = null;
			if(this.cancelEvent)
			{
				global.element.removeEventListener("click", this.cancelEvent, false);
				this.cancelEvent = null;
			}
			
			this.arrow.className = "combobox-arrow-disabled";
			this.arrow.childNodes[0].className = jaswin.ICON_TRIANGLE_BOTTOM_DISABLED;
			if(this.arrowClickEvent)
			{
				this.arrow.removeEventListener("click", this.arrowClickEvent, false);
				this.arrowClickEvent = null;
			}
			
			if(this.style == list.STYLE_EDITABLE)
			{
				this.text.disabled = true;
				if(this.focusEvent)
				{
					this.text.removeEventListener("focus", this.focusEvent, false);
					this.focusEvent = null;
				}
				if(this.blurEvent)
				{
					this.text.removeEventListener("blur", this.blurEvent, false);
					this.blurEvent = null;
				}
				if(this.inputEvent)
				{
					this.text.removeEventListener("input", this.inputEvent, false);
					this.inputEvent = null;
				}
				if(this.comboArrowDown)
				{
					this.text.removeEventListener("mousedown", this.comboArrowDown, false);
					this.comboArrowDown = null;
				}
			}
			else
			{
				this.text.className = "combobox-label-disabled";
				if(this.clickEvent)
				{
					this.text.removeEventListener("click", this.clickEvent, false);
					this.clickEvent = null;
				}
				if(this.focusEvent)
				{
					this.element.removeEventListener("focus", this.focusEvent, false);
					this.focusEvent = null;
				}
				if(this.blurEvent)
				{
					this.element.removeEventListener("blur", this.blurEvent, false);
					this.blurEvent = null;
				}
			}
			if(this.mouseMoveEvent)
			{
				this.element.removeEventListener("mousemove", this.mouseMoveEvent, false);
				this.mouseMoveEvent = null;
			}
			if(this.mouseOutEvent)
			{
				this.element.removeEventListener("mouseout", this.mouseOutEvent, false);
				this.mouseOutEvent = null;
			}
			if(this.keyDownEvent)
			{
				this.element.removeEventListener("keydown", this.keyDownEvent, false);
				this.keyDownEvent = null;
			}
			if(this.mouseDownEvent)
			{
				this.list.removeEventListener("mousedown", this.mouseDownEvent, false);
				this.mouseDownEvent = null;
			}
			if(this.mouseUpEvent)
			{
				this.list.removeEventListener("mouseup", this.mouseUpEvent, false);
				this.mouseUpEvent = null;
			}
		}
		
		//==================================================
		//	Set this combobox hover
		//==================================================
		setHover()
		{
			this.box.className = "combobox-focus";
			this.arrow.className = "combobox-arrow-focus";
		}
		
		//==================================================
		//	Set this combobox mouse out
		//==================================================
		setOut()
		{
			if(this.style == list.STYLE_EDITABLE)
			{
				if(document.activeElement == this.text)
					return;
			}
			
			if(!this.dropDown)
			{
				this.box.className = "combobox";
				this.arrow.className = "combobox-arrow";
			}
		}
		
		//==================================================
		//	Blur from this combobox
		//==================================================
		blur()
		{
			if(this.dropDown)
			{
				this.list.style.display = "none";
				this.dropDown = false;
				oDropDown = null;
				if(this.cancelEvent)
				{
					global.element.removeEventListener("click", this.cancelEvent, false);
					this.cancelEvent = null;
				}
			}
			this.box.className = "combobox";
			this.arrow.className = "combobox-arrow";
			
			this.currentIndex = -1;
			this.setFocusItem(-1);
		}
		
		//==================================================
		//	Click this combobox
		//==================================================
		click()
		{
			if(this.dropDown)
			{
				// Dropped down to hidden
				this.list.style.display = "none";
				this.dropDown = false;
				oDropDown = null;
				if(this.cancelEvent)
				{
					global.element.removeEventListener("click", this.cancelEvent, false);
					this.cancelEvent = null;
				}
				
				this.box.className = "combobox";
				this.arrow.className = "combobox-arrow";
				this.currentIndex = -1;
				
				let iCounter;
				for(iCounter = 0; iCounter < this.items.length; iCounter++)
				{
					if(!this.items[iCounter].selected)
					{
						this.items[iCounter].element.className = "list-item";
						if(this.style == list.STYLE_CHECKBOX)
							this.items[iCounter].box.className = jaswin.ICON_CHECKBOX;
					}
				}
				this.list.scrollTop = 0;
			}
			else
			{
				// Hidden to dropped down
				let oWindow = jaswin.findParentWindow(this);
				let iListHeight;
				let bBellow = true;
				let oThisRect;
				let oParentRect;
				
				// Which show list bellow or above?
				if(this.items.length >= this.maxDropdown)
					iListHeight = 20 * this.maxDropdown;
				else
					iListHeight = 20 * this.items.length;
				oThisRect = this.element.getBoundingClientRect();
				if(oWindow == global)
				{
					oParentRect = global.container.getBoundingClientRect();
					if(oThisRect.top - oParentRect.top < iListHeight)
						bBellow = true;
					else if(oThisRect.top + 20 + iListHeight + 4 > oParentRect.bottom)	// 4:border of box and list
						bBellow = false;
				}
				else
				{
					oParentRect = oWindow.parent.container.getBoundingClientRect();
					if(oThisRect.top - oParentRect.top < iListHeight)
						bBellow = true;
					else if(oThisRect.top + 20 + iListHeight + 4 > oParentRect.bottom)	// 4:border of box and list
						bBellow = false;
				}
				if(bBellow)
					this.list.style.top = "20px";
				else
					this.list.style.top = (1 + iListHeight) * (-1) + "px";
				
				if((jaswin.frame) && (jaswin.frame.windowCount))
				{
					if(oWindow != global)
						this.list.style.zIndex = 100 + jaswin.frame.windowCount + oWindow.modalLevel;
					else
						this.list.style.zIndex = 100 + jaswin.frame.windowCount;
				}
				else
					this.list.style.zIndex = 100;
				this.list.style.display = "block";
				this.dropDown = true;
				oDropDown = this;
				
				if(this.style == list.STYLE_CHECKBOX)
				{
					if(this.selectedIndexes.length > 0)
						this.currentIndex = this.selectedIndexes[0];
					else
						this.currentIndex = -1;
				}
				else
					this.currentIndex = this.selectedIndex;
				
				if(this.currentIndex < 0)
					this.list.scrollTop = 0;
				else
				{
					if(this.currentIndex * 20 < this.list.scrollTop)
						this.list.scrollTop = this.currentIndex * 20;
					else if((this.currentIndex + 1) * 20 > this.maxDropdown * 20 + this.list.scrollTop)
						this.list.scrollTop = (this.currentIndex + 1) * 20 - (this.maxDropdown * 20 + this.list.scrollTop);
				}
				
				if(this.style == list.STYLE_EDITABLE)
					this.text.focus();
				
				if(!this.cancelEvent)
				{
					global.element.addEventListener("click", this.cancelEvent = (e) => {
						if(!oDropDown)
							return;
						
						let oElement = e.target;
						if(!oDropDown.element.contains(oElement))
							oDropDown.click();
					}, false);
				}
			}
		}
		
		//==================================================
		//	Focus textbox of this combobox
		//==================================================
		focusText()
		{
			if(this.style != list.STYLE_EDITABLE)
				return;
			
			this.text.select();
			this.setHover();
		}
		
		//==================================================
		//	Blur textbox of this combobox
		//==================================================
		blurText()
		{
			if(this.style != list.STYLE_EDITABLE)
				return;
			
			if(this == oCurrentMouseDown)
				return;
			
			if(this.closingList)
			{
				this.closingList = false;
				return;
			}
			
			this.list.style.display = "none";
			this.dropDown = false;
			oDropDown = null;
			if(this.cancelEvent)
			{
				global.element.removeEventListener("click", this.cancelEvent, false);
				this.cancelEvent = null;
			}
			
			this.box.className = "combobox";
			this.arrow.className = "combobox-arrow";
			
			this.setOut();
			this.currentIndex = -1;
			this.setFocusItem(-1);
		}
		
		//==================================================
		//	Input textbox of this combobox
		//==================================================
		inputText()
		{
			if(this.style != list.STYLE_EDITABLE)
				return;
			
			if(this.text.value == "")
			{
				this.selectList(-1, false);
				this.list.style.display = "none";
				this.dropDown = false;
				oDropDown = null;
				this.list.scrollTop = 0;
				this.currentIndex = -1;
				if(this.cancelEvent)
				{
					global.element.removeEventListener("click", this.cancelEvent, false);
					this.cancelEvent = null;
				}
				return;
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.text.value == this.items[iCounter].text)
				{
					this.selectList(iCounter, true);
					this.list.scrollTop = 20 * iCounter;
					return;
				}
			}
			
			this.selectList(-1, false);
			this.currentIndex = -1;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter].text.indexOf(this.text.value) == 0)
				{
					if(iCounter * 20 < this.list.scrollTop)
						this.list.scrollTop = this.iCounter * 20;
					else if((iCounter + 1) * 20 > this.maxDropdown * 20 + this.list.scrollTop)
						this.list.scrollTop = (iCounter + 1) * 20 - (this.maxDropdown * 20 + this.list.scrollTop);
					return;
				}
			}
		}
		
		//==================================================
		//	Add item to this combobox
		//==================================================
		addItem(sText, sVal, bSelected)
		{
			if(typeof sText != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, this.id),
											"list.js", "jaswin.list.ComboBox.addItem");
			
			if(typeof sVal != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
											"list.js", "jaswin.list.ComboBox.addItem");
			
			if(typeof bSelected != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SELECTED, this.id),
											"list.js", "jaswin.list.ComboBox.addItem");
			
			let iWidth;
			let iCounter;
			let oItem;
			let bEnabled = false;
			
			// Adjust list width and height
			iWidth = this.rect.width - 2 - (1 * 2) - (1 * 2);	// 2:padding 1*2:border(frame) 1*2:border(item)
			if(this.style == list.STYLE_CHECKBOX)
				iWidth -= 17;									// 17:checkbox
			
			if(this.items.length + 1 <= this.maxDropdown)
			{
				// Height of list
				this.list.style.height = (this.items.length + 1) * 20 + "px";
			}
			else
			{
				// Width of label
				iWidth -= 17;									// 17:scrollbar
				for(iCounter = 0; iCounter < this.items.length; iCounter++)
					this.items[iCounter].label.style.width = iWidth + "px";
			}
			
			// Make a new item
			if((this.enabled) && (this.parentEnabled))
				bEnabled = true;
			oItem = new list.ListItem(this, this.style, iWidth, this.align,
										sText, sVal, bEnabled, this.draggable, this.onDragstart);
			
			this.list.appendChild(oItem.element);
			this.items.push(oItem);
			
			if(bSelected)
				this.selectList(this.items.length - 1, false);
		}
		
		//==================================================
		//	Remove item from this combobox
		//==================================================
		removeItem(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"list.js", "jaswin.list.ComboBox.removeItem");
			if((iIndex < 0) || (iIndex >= this.items.length))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"list.js", "jaswin.list.ComboBox.removeItem");
			// Remove an item
			this.list.removeChild(this.items[iIndex].element);
			this.items.splice(iIndex, 1);
			
			// Adjust list width and height
			iWidth = this.rect.width - 2 - (1 * 2) - (1 * 2);	// 2:padding 1*2:border(frame) 1*2:border(item)
			if(this.style == list.STYLE_CHECKBOX)
				iWidth -= 17;									// 17:checkbox
			
			if(this.items.length <= this.maxDropdown)
				this.list.style.height = this.items.length * 20 + "px";
			else
			{
				this.list.style.height = this.maxDropdown * 20 + "px";
				iWidth -= 17;									// 17:scrollbar
			}
			
			// Width of label and selection
			let sLabelText = "";
			this.selectedIndexes = new Array();
			this.selectedIndex = -1;
			this.currentIndex = -1;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				this.items[iCounter].label.style.width = iWidth + "px";
				if(this.items[iCounter].selected)
				{
					if(this.style == list.STYLE_CHECKBOX)
					{
						this.selectedIndexes.push(iCounter);
						if(sLabelText.length > 0)
							sLabelText += ",";
						sLabelText += this.items[iCounter].text;
					}
					else
					{
						this.selectedIndexes[0] = iCounter;
						sLabelText = this.items[iCounter].text;
					}
					this.selectedIndex = iCounter;
				}
			}
			if(this.style == list.STYLE_EDITABLE)
				this.text.value = sLabelText;
			else
				this.text.innerHTML = sLabelText;
		}
		
		//==================================================
		//	Set list item of this combobox hover
		//==================================================
		setHoverList(oItem)
		{
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
				{
					this.currentIndex = iCounter;
					if(!oItem.selected)
					{
						oItem.element.className = "list-item-hover";
						if(this.style == list.STYLE_CHECKBOX)
							oItem.box.className = jaswin.ICON_CHECKBOX_HOVER;
					}
					break;
				}
			}
		}
		
		//==================================================
		//	Set list item of this combobox mouse out
		//==================================================
		setOutList(oItem)
		{
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
				{
					this.currentIndex = -1;
					if(!oItem.selected)
					{
						oItem.element.className = "list-item";
						if(this.style == list.STYLE_CHECKBOX)
							oItem.box.className = jaswin.ICON_CHECKBOX;
					}
					break;
				}
			}
		}
		
		//==================================================
		//	Click list item of this combobox
		//==================================================
		clickList(oItem)
		{
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
				{
					this.selectList(iCounter, true);
					break;
				}
			}
			
			if(this.style == list.STYLE_CHECKBOX)
				bDropDownKeep = true;
			else
			{
				this.list.style.display = "none";
				this.dropDown = false;
				oDropDown = null;
				if(this.cancelEvent)
				{
					global.element.removeEventListener("click", this.cancelEvent, false);
					this.cancelEvent = null;
				}
				
				this.box.className = "combobox";
				this.arrow.className = "combobox-arrow";
			}
		}
		
		//==================================================
		//	Set list item of this combobox focused
		//==================================================
		setFocusItem(iIndex)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"list.js", "jaswin.list.ComboBox.setFocusItem");
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(iCounter == iIndex)
				{
					this.currentIndex = iCounter;
					if(!this.items[iCounter].selected)
					{
						this.items[iCounter].element.className = "list-item-hover";
						if(this.style == list.STYLE_CHECKBOX)
							this.items[iCounter].box.className = jaswin.ICON_CHECKBOX_HOVER;
					}
				}
				else
				{
					if(!this.items[iCounter].selected)
					{
						this.items[iCounter].element.className = "list-item";
						if(this.style == list.STYLE_CHECKBOX)
							this.items[iCounter].box.className = jaswin.ICON_CHECKBOX;
					}
				}
			}
			
			if(iIndex == -1)
				this.list.scrollTop = 0;
			else
			{
				if(iIndex * 20 < this.list.scrollTop)
					this.list.scrollTop = iIndex * 20;
				else if((iIndex + 1) * 20 > this.maxDropdown * 20 + this.list.scrollTop)
					this.list.scrollTop = (iIndex + 1) * 20 - (this.maxDropdown * 20 + this.list.scrollTop);
			}
		}
		
		//==================================================
		//	Select list item of this combobox
		//==================================================
		selectList(iIndex, bExecFunc)
		{
			if(typeof iIndex != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_INDEX, this.id),
											"list.js", "jaswin.list.ComboBox.selectList");
			if(iIndex >= this.items.length)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_INDEX, this.id),
											"list.js", "jaswin.list.ComboBox.selectList");
			
			let bExec = true;
			
			if((typeof bExecFunc != "undefined") && (bExecFunc != null))
			{
				if(typeof bExecFunc != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_EXEC_FUNC, this.id),
												"list.js", "jaswin.list.ComboBox.selectList");
				bExec = bExecFunc;
			}
			
			if((this.style != list.STYLE_CHECKBOX)
			&& (iIndex == this.selectedIndex))
				return;
			
			let iCounter;
			let sLabelText = "";
			
			this.selectedIndexes = new Array();
			
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.style == list.STYLE_CHECKBOX)
				{
					if(iCounter == iIndex)
					{
						if(this.items[iIndex].selected)
						{
							if((this.enabled) && (this.parentEnabled))
							{
								this.items[iIndex].element.className = "list-item";
								this.items[iIndex].box.className = jaswin.ICON_CHECKBOX;
							}
							this.items[iIndex].selected = false;
						}
						else
						{
							if((this.enabled) && (this.parentEnabled))
							{
								this.items[iIndex].element.className = "list-item-selected";
								this.items[iIndex].box.className = jaswin.ICON_CHECKBOX_CHECKED;
							}
							this.items[iIndex].selected = true;
							this.selectedIndexes.push(iIndex);
							
							if(sLabelText.length > 0)
								sLabelText += ",";
							sLabelText += this.items[iIndex].text;
						}
					}
					else if(this.items[iCounter].selected)
					{
						this.selectedIndexes.push(iCounter);
						if(sLabelText.length > 0)
							sLabelText += ",";
						sLabelText += this.items[iCounter].text;
					}
				}
				else
				{
					if(iCounter == iIndex)
					{
						if((this.enabled) && (this.parentEnabled))
							this.items[iIndex].element.className = "list-item-selected";
						this.items[iIndex].selected = true;
						this.selectedIndex = iIndex;
						sLabelText = this.items[iIndex].text;
					}
					else
					{
						if((this.enabled) && (this.parentEnabled))
							this.items[iCounter].element.className = "list-item";
						this.items[iCounter].selected = false;
					}
				}
			}
			
			if(this.style != list.STYLE_CHECKBOX)
			{
				this.selectedIndex = iIndex;
				this.selectedIndexes[0] = iIndex;
				this.currentIndex = iIndex;
			}
			
			if(this.style == list.STYLE_EDITABLE)
			{
				if((iIndex != -1)
				&& (this.text.value != sLabelText))
					this.text.value = sLabelText;
			}
			else
				this.text.innerHTML = sLabelText;
			
			if((this.onChange) && (this.enabled) && (this.parentEnabled)
			&& (bExec) && (iIndex != -1))
				this.onChange(this.items[iIndex]);
		}
		
		//==================================================
		//	Select list item of this combobox by value
		//==================================================
		selectListByValue(sValue, bExecFunc)
		{
			if(typeof sValue != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
											"list.js", "jaswin.list.ComboBox.selectListByValue");
			
			let bExec = true;
			
			if((typeof bExecFunc != "undefined") && (bExecFunc != null))
			{
				if(typeof bExecFunc != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_EXEC_FUNC, this.id),
												"list.js", "jaswin.list.ComboBox.selectListByValue");
				bExec = bExecFunc;
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter].value == sValue)
				{
					this.selectList(iCounter, bExec);
					return;
				}
			}
			
			this.selectList(-1, bExec);
		}
		
		//==================================================
		//	Key down on this combobox
		//==================================================
		keyDownCombo(e)
		{
			if(e.keyCode == 13)
			{
				// Enter
				if(this.currentIndex != -1)
				{
					if(this.style != list.STYLE_CHECKBOX)
						this.selectList(this.currentIndex, true);
					this.click();
					this.setHover();
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			}
			else if(e.keyCode == 27)
			{
				// Esc
				if(this.dropDown)
				{
					this.click();
					this.setHover();
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			}
			else if(e.keyCode == 32)
			{
				// Space
				if(this.style != list.STYLE_EDITABLE)
				{
					if(this.style == list.STYLE_CHECKBOX)
					{
						if(!this.dropDown)
							this.click();
						else
							this.selectList(this.currentIndex, true);
					}
					else
					{
						if((oDropDown) && (oDropDown != this))
							oDropDown.click();
						this.click();
						this.setHover();
					}
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			}
			else if(e.keyCode == 40)
			{
				// down
				if(!this.dropDown)
				{
					this.click();
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				if(this.currentIndex < this.items.length - 1)
					this.setFocusItem(this.currentIndex + 1);
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			else if(e.keyCode == 38)
			{
				// up
				if(this.currentIndex <= 0)
				{
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				if(this.dropDown)
				{
					this.setFocusItem(this.currentIndex - 1);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			}
			else if((e.keyCode == 9)
				&& (this.style == list.STYLE_EDITABLE))
			{
				this.text.selectionStart = 0;
				this.text.selectionEnd = 0;
				window.getSelection().removeAllRanges();
			}
			return true;
		}
		
		//==================================================
		//	Clear list of this combobox
		//==================================================
		clear()
		{
			this.currentIndex = -1;
			this.selectedIndex = -1;
			this.selectedIndexes = new Array();
			this.items = new Array();
			this.list.innerHTML = "";
			this.list.style.height = "0px";
			if(this.style == list.STYLE_EDITABLE)
				this.text.value = "";
			else
				this.text.innerHTML = "";
		}
		
		//==================================================
		//	Get selected value
		//==================================================
		selectedValue()
		{
			if(this.style == list.STYLE_CHECKBOX)
			{
				if(this.selectedIndexes.length == 0)
					return null;
				
				var sRet = "";
				var iCounter;
				
				for(iCounter = 0; iCounter < this.selectedIndexes.length; iCounter++)
				{
					if(iCounter != 0)
						sRet += ",";
					sRet += this.items[this.selectedIndexes[iCounter]].value;
				}
				
				return sRet;
			}
			else
			{
				if(this.selectedIndex == -1)
					return null;
				
				return this.items[this.selectedIndex].value;
			}
		}
		
		//==================================================
		//	Get selected text
		//==================================================
		selectedText()
		{
			if(this.style == list.STYLE_EDITABLE)
				return this.text.value;
			
			if(this.style == list.STYLE_CHECKBOX)
			{
				if(this.selectedIndexes.length == 0)
					return null;
				else
					return this.text.innerHTML;
			}
			else
			{
				if(this.selectedIndex == -1)
					return null;
				else
					return this.text.innerHTML;
			}
		}
		
		//==================================================
		//	Get selected items
		//==================================================
		getSelectedItems()
		{
			let oaItems;
			let iCounter;
			
			oaItems = new Array();
			for(iCounter = 0; iCounter < this.selectedIndexes.length; iCounter++)
				oaItems.push(this.items[this.selectedIndexes[iCounter]]);
			
			return oaItems;
		}
		
		//==================================================
		//	Get text by value
		//==================================================
		getTextByValue(sValue)
		{
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter].value == sValue)
					return this.items[iCounter].text;
			}
			
			return null;
		}
		
		//==================================================
		//	Contain item or not
		//==================================================
		contains(oItem)
		{
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
			{
				if(this.items[iCounter] == oItem)
					return true;
			}
			
			return false;
		}
		
		//==================================================
		//	Sort
		//==================================================
		sort(iSortBy, iAscDesc)
		{
			if(typeof iSortBy != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SORT_BY, this.id),
											"list.js", "jaswin.list.ComboBox.sort");
			if(typeof iAscDesc != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SORT_TYPE, this.id),
											"list.js", "jaswin.list.ComboBox.sort");
			if(!this.items.length)
				return;
			
			// Execute sort
			if(iSortBy == list.SORT_BY_TEXT)
			{
				if(iAscDesc == list.SORT_ASC)
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.text < oItemB.text) return -1;
										if(oItemA.text > oItemB.text) return 1;
										return 0;
									});
				}
				else
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.text > oItemB.text) return -1;
										if(oItemA.text < oItemB.text) return 1;
										return 0;
									});
				}
			}
			else
			{
				if(iAscDesc == list.SORT_ASC)
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.value < oItemB.value) return -1;
										if(oItemA.value > oItemB.value) return 1;
										return 0;
									});
				}
				else
				{
					this.items.sort(function(oItemA, oItemB){
										if(oItemA.value > oItemB.value) return -1;
										if(oItemA.value < oItemB.value) return 1;
										return 0;
									});
				}
			}
			
			// Replace elements
			let iCounter;
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
				this.items[iCounter].element.parentNode.removeChild(this.items[iCounter].element);
			for(iCounter = 0; iCounter < this.items.length; iCounter++)
				this.list.appendChild(this.items[iCounter].element);
		}
		
		//==================================================
		//	Set tooltip
		//==================================================
		setTooltip(sTooltip)
		{
			if(typeof sTooltip != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, this.id),
											"list.js", "jaswin.list.ComboBox.setTooltip");
			this.element.title = sTooltip;
		}
		
		//==================================================
		//	Set change event
		//==================================================
		setChangeEvent(fpChange)
		{
			if(typeof fpChange != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, this.id),
											"list.js", "jaswin.list.ComboBox.setChangeEvent");
			
			this.onChange = fpChange;
		}
	}
	
}(jaswin.list = jaswin.list || {}));
